@@ -1,13 +1,13 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2017 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 
-from xpra.util import nonl, csv, std, envbool, print_nested_dict, repr_ellipsized
+from xpra.util import nonl, csv, std, envbool, print_nested_dict, ellipsizer
 from xpra.os_util import POSIX, bytestostr
 from xpra.log import Logger
 
@@ -16,7 +16,7 @@ log = Logger("keyboard")
 LAYOUT_GROUPS = envbool("XPRA_LAYOUT_GROUPS", True)
 
 
-class KeyboardHelper(object):
+class KeyboardHelper:
 
     def __init__(self, net_send, keyboard_sync=True,
                  shortcut_modifiers="auto", key_shortcuts=(),
@@ -37,7 +37,7 @@ class KeyboardHelper(object):
         self.options = options
         #the platform class which allows us to map the keys:
         from xpra.platform.keyboard import Keyboard
-        self.keyboard = Keyboard()
+        self.keyboard = Keyboard()      #pylint: disable=not-callable
         log("KeyboardHelper(%s) keyboard=%s",
             (net_send, keyboard_sync, key_shortcuts,
              raw, layout, layouts, variant, variants, options), self.keyboard)
@@ -86,7 +86,6 @@ class KeyboardHelper(object):
 
 
     def parse_shortcuts(self, strs):
-        #TODO: maybe parse with re instead?
         if not strs:
             """ if none are defined, add this as default
             it would be nicer to specify it via OptionParser in main
@@ -210,7 +209,7 @@ class KeyboardHelper(object):
                     modifiers.append(imod)
                 if not valid:
                     continue
-            #TODO: validate keyname
+            #should we be validating the keyname?
             keyname = keyspec[len(keyspec)-1]
             shortcuts.setdefault(keyname, []).append((modifiers, action, args))
             log("shortcut(%s)=%s", s, csv((modifiers, action, args)))
@@ -221,7 +220,7 @@ class KeyboardHelper(object):
     def key_handled_as_shortcut(self, window, key_name, modifiers, depressed):
         #find the shortcuts that may match this key:
         shortcuts = self.key_shortcuts.get(key_name)
-        log("key_handled_as_shortcut: shortcut(%s)=%s", key_name, shortcuts)
+        log("key_handled_as_shortcut%s shortcuts=%s", (window, key_name, modifiers, depressed), shortcuts)
         if not shortcuts:
             return False
         if len(shortcuts)>1:
@@ -242,6 +241,8 @@ class KeyboardHelper(object):
         for rm in req_mods:
             if rm not in modifiers:
                 #modifier is missing, bail out
+                log("not matched %s for %s: %s not in %s",
+                    shortcut, key_name, rm, modifiers)
                 return False
             try:
                 extra_modifiers.remove(rm)
@@ -345,25 +346,32 @@ class KeyboardHelper(object):
         return _print, query, query_struct
 
     def query_xkbmap(self):
+        log("query_xkbmap()")
         (
             self.xkbmap_layout, self.xkbmap_layouts,
             self.xkbmap_variant, self.xkbmap_variants,
             self.xkbmap_options,
             ) = self.get_layout_spec()
-        self.xkbmap_print, self.xkbmap_query, self.xkbmap_query_struct = self.get_keymap_spec()
+        spec = self.get_keymap_spec()
+        self.xkbmap_print, self.xkbmap_query, self.xkbmap_query_struct = spec
+        log("query_xkbmap() get_keymap_spec()=%s", spec)
         self.xkbmap_keycodes = self.get_full_keymap()
+        log("query_xkbmap() get_full_keymap()=%s", self.xkbmap_keycodes)
         self.xkbmap_x11_keycodes = self.keyboard.get_x11_keymap()
+        log("query_xkbmap() %s.get_x11_keymap()=%s", self.keyboard, self.xkbmap_x11_keycodes)
+        mods = self.keyboard.get_keymap_modifiers()
         (
             self.xkbmap_mod_meanings,
             self.xkbmap_mod_managed,
             self.xkbmap_mod_pointermissing,
-            ) = self.keyboard.get_keymap_modifiers()
+            ) = mods
+        log("query_xkbmap() %s.get_keymap_modifiers()=%s", self.keyboard, mods)
         self.update_hash()
         log("layout=%s, layouts=%s, variant=%s, variants=%s",
             self.xkbmap_layout, self.xkbmap_layouts, self.xkbmap_variant, self.xkbmap_variants)
         log("print=%s, query=%s, struct=%s", nonl(self.xkbmap_print), nonl(self.xkbmap_query), self.xkbmap_query_struct)
-        log("keycodes=%s", repr_ellipsized(str(self.xkbmap_keycodes)))
-        log("x11 keycodes=%s", repr_ellipsized(str(self.xkbmap_x11_keycodes)))
+        log("keycodes=%s", ellipsizer(self.xkbmap_keycodes))
+        log("x11 keycodes=%s", ellipsizer(self.xkbmap_x11_keycodes))
         log("mod managed: %s", self.xkbmap_mod_managed)
         log("mod meanings: %s", self.xkbmap_mod_meanings)
         log("mod pointermissing: %s", self.xkbmap_mod_pointermissing)
@@ -430,11 +438,11 @@ class KeyboardHelper(object):
         kb_info = {}
         if self.xkbmap_query_struct or self.xkbmap_query:
             xkbqs = self.xkbmap_query_struct
-            if xkbqs:
+            if not xkbqs:
                 #parse query into a dict
                 from xpra.keyboard.layouts import parse_xkbmap_query
                 xkbqs = parse_xkbmap_query(self.xkbmap_query)
-            for x in ["rules", "model", "layout"]:
+            for x in ("rules", "model", "layout"):
                 v = xkbqs.get(x)
                 if v:
                     kb_info[x] = v
