@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -9,16 +9,16 @@ import subprocess
 import hashlib
 import uuid
 
-from xpra.log import Logger
-printlog = Logger("printing")
-filelog = Logger("file")
-
 from xpra.child_reaper import getChildReaper
-from xpra.os_util import monotonic_time, bytestostr, strtobytes, POSIX, WIN32
+from xpra.os_util import monotonic_time, bytestostr, strtobytes, umask_context, POSIX, WIN32
 from xpra.util import typedict, csv, nonl, envint, envbool, engs
 from xpra.scripts.config import parse_bool
 from xpra.simple_stats import std_unit
 from xpra.make_thread import start_thread
+from xpra.log import Logger
+
+printlog = Logger("printing")
+filelog = Logger("file")
 
 DELETE_PRINTER_FILE = envbool("XPRA_DELETE_PRINTER_FILE", True)
 FILE_CHUNKS_SIZE = max(0, envint("XPRA_FILE_CHUNKS_SIZE", 65536))
@@ -37,48 +37,73 @@ DENY = 0
 ACCEPT = 1
 OPEN = 2
 
+def basename(filename):
+    #we can't use os.path.basename,
+    #because the remote end may have sent us a filename
+    #which is using a different pathsep
+    tmp = filename
+    for sep in ("\\", "/", os.sep):
+        i = tmp.rfind(sep) + 1
+        tmp = tmp[i:]
+    filename = tmp
+    if WIN32:   # pragma: no cover
+        #many characters aren't allowed at all on win32:
+        tmp = ""
+        for char in filename:
+            if ord(char)<32 or char in ("<", ">", ":", "\"", "|", "?", "*"):
+                char = "_"
+            tmp += char
+    return tmp
 
 def safe_open_download_file(basefilename, mimetype):
     from xpra.platform.paths import get_download_dir
-    #make sure we use a filename that does not exist already:
     dd = os.path.expanduser(get_download_dir())
-    base = os.path.basename(basefilename)
-    wanted_filename = os.path.abspath(os.path.join(dd, base))
+    filename = os.path.abspath(os.path.join(dd, basename(basefilename)))
     ext = MIMETYPE_EXTS.get(mimetype)
-    if ext:
+    if ext and not filename.endswith("."+ext):
         #on some platforms (win32),
         #we want to force an extension
         #so that the file manager can display them properly when you double-click on them
-        if not wanted_filename.endswith("."+ext):
-            wanted_filename += "."+ext
-    filename = wanted_filename
+        filename += "."+ext
+    #make sure we use a filename that does not exist already:
+    root, ext = os.path.splitext(filename)
     base = 0
     while os.path.exists(filename):
         filelog("cannot save file as %s: file already exists", filename)
-        root, ext = os.path.splitext(wanted_filename)
         base += 1
         filename = root+("-%s" % base)+ext
+    filelog("safe_open_download_file(%s, %s) will use '%s'", basefilename, mimetype, filename)
     flags = os.O_CREAT | os.O_RDWR | os.O_EXCL
     try:
         flags |= os.O_BINARY                #@UndefinedVariable (win32 only)
-    except:
+    except AttributeError:
         pass
-    fd = os.open(filename, flags)
-    filelog("using filename '%s'", filename)
+    with umask_context(0o133):
+        fd = os.open(filename, flags)
+    filelog("using filename '%s', file descriptor=%s", filename, fd)
     return filename, fd
 
+def s(v):
+    try:
+        return v.decode("utf8")
+    except:
+        return bytestostr(v)
 
-class FileTransferAttributes(object):
+
+class FileTransferAttributes:
 
     def __init__(self):
         self.init_attributes()
 
     def init_opts(self, opts, can_ask=True):
         #get the settings from a config object
-        self.init_attributes(opts.file_transfer, opts.file_size_limit, opts.printing, opts.open_files, opts.open_url, opts.open_command, can_ask)
+        self.init_attributes(opts.file_transfer, opts.file_size_limit,
+                             opts.printing, opts.open_files, opts.open_url, opts.open_command, can_ask)
 
-    def init_attributes(self, file_transfer="yes", file_size_limit=10, printing="yes", open_files="no", open_url="yes", open_command=None, can_ask=True):
-        filelog("file transfer: init_attributes%s", (file_transfer, file_size_limit, printing, open_files, open_url, open_command, can_ask))
+    def init_attributes(self, file_transfer="no", file_size_limit=10, printing="no",
+                        open_files="no", open_url="no", open_command=None, can_ask=True):
+        filelog("file transfer: init_attributes%s",
+                (file_transfer, file_size_limit, printing, open_files, open_url, open_command, can_ask))
         def pbool(name, v):
             return parse_bool(name, v, True)
         def pask(v):
@@ -102,7 +127,7 @@ class FileTransferAttributes(object):
         self.open_command = open_command
         filelog("file transfer attributes=%s", self.get_file_transfer_features())
 
-    def get_file_transfer_features(self):
+    def get_file_transfer_features(self) -> dict:
         #used in hello packets
         return {
                 "file-transfer"     : self.file_transfer,
@@ -118,7 +143,7 @@ class FileTransferAttributes(object):
                 "file-ask-timeout"  : self.file_ask_timeout,
                 }
 
-    def get_info(self):
+    def get_info(self) -> dict:
         #slightly different from above... for legacy reasons
         #this one is used for get_info() in a proper "file." namespace from server_base.py
         return {
@@ -143,7 +168,7 @@ class FileTransferHandler(FileTransferAttributes):
     """
 
     def init_attributes(self, *args):
-        FileTransferAttributes.init_attributes(self, *args)
+        super().init_attributes(*args)
         self.remote_file_transfer = False
         self.remote_file_transfer_ask = False
         self.remote_printing = False
@@ -161,11 +186,10 @@ class FileTransferHandler(FileTransferAttributes):
         self.receive_chunks_in_progress = {}
         self.file_descriptors = set()
         if not getattr(self, "timeout_add", None):
-            from xpra.gtk_common.gobject_compat import import_glib
-            glib = import_glib()
-            self.timeout_add = glib.timeout_add
-            self.idle_add = glib.idle_add
-            self.source_remove = glib.source_remove
+            from gi.repository import GLib
+            self.timeout_add = GLib.timeout_add
+            self.idle_add = GLib.idle_add
+            self.source_remove = GLib.source_remove
 
     def cleanup(self):
         for t in self.pending_send_data_timers.values():
@@ -177,8 +201,8 @@ class FileTransferHandler(FileTransferAttributes):
         self.receive_chunks_in_progress = {}
         for x in tuple(self.file_descriptors):
             try:
-                x.close()
-            except:
+                os.close(x)
+            except OSError:
                 pass
         self.file_descriptors = set()
         self.init_attributes()
@@ -198,26 +222,30 @@ class FileTransferHandler(FileTransferAttributes):
         self.dump_remote_caps()
 
     def dump_remote_caps(self):
-        filelog("file transfer remote caps: file-transfer=%s   (ask=%s)", self.remote_file_transfer, self.remote_file_transfer_ask)
-        filelog("file transfer remote caps: printing=%s        (ask=%s)", self.remote_printing, self.remote_printing_ask)
-        filelog("file transfer remote caps: open-files=%s      (ask=%s)", self.remote_open_files, self.remote_open_files_ask)
-        filelog("file transfer remote caps: open-url=%s        (ask=%s)", self.remote_open_url, self.remote_open_url_ask)
+        filelog("file transfer remote caps: file-transfer=%s   (ask=%s)",
+                self.remote_file_transfer, self.remote_file_transfer_ask)
+        filelog("file transfer remote caps: printing=%s        (ask=%s)",
+                self.remote_printing, self.remote_printing_ask)
+        filelog("file transfer remote caps: open-files=%s      (ask=%s)",
+                self.remote_open_files, self.remote_open_files_ask)
+        filelog("file transfer remote caps: open-url=%s        (ask=%s)",
+                self.remote_open_url, self.remote_open_url_ask)
 
-    def get_info(self):
+    def get_info(self) -> dict:
         info = FileTransferAttributes.get_info(self)
         info["remote"] = {
-                          "file-transfer"   : self.remote_file_transfer,
-                          "file-transfer-ask" : self.remote_file_transfer_ask,
-                          "file-size-limit" : self.remote_file_size_limit,
-                          "file-chunks"     : self.remote_file_chunks,
-                          "open-files"      : self.remote_open_files,
-                          "open-files-ask"  : self.remote_open_files_ask,
-                          "open-url"        : self.remote_open_url,
-                          "open-url-ask"    : self.remote_open_url_ask,
-                          "printing"        : self.remote_printing,
-                          "printing-ask"    : self.remote_printing_ask,
-                          "file-ask-timeout" : self.remote_file_ask_timeout,
-                          }
+            "file-transfer"     : self.remote_file_transfer,
+            "file-transfer-ask" : self.remote_file_transfer_ask,
+            "file-size-limit"   : self.remote_file_size_limit,
+            "file-chunks"       : self.remote_file_chunks,
+            "open-files"        : self.remote_open_files,
+            "open-files-ask"    : self.remote_open_files_ask,
+            "open-url"          : self.remote_open_url,
+            "open-url-ask"      : self.remote_open_url_ask,
+            "printing"          : self.remote_printing,
+            "printing-ask"      : self.remote_printing_ask,
+            "file-ask-timeout"  : self.remote_file_ask_timeout,
+            }
         return info
 
     def check_digest(self, filename, digest, expected_digest, algo="sha1"):
@@ -240,10 +268,11 @@ class FileTransferHandler(FileTransferAttributes):
 
     def _process_send_file_chunk(self, packet):
         chunk_id, chunk, file_data, has_more = packet[1:5]
+        chunk_id = bytestostr(chunk_id)
         filelog("_process_send_file_chunk%s", (chunk_id, chunk, "%i bytes" % len(file_data), has_more))
         chunk_state = self.receive_chunks_in_progress.get(chunk_id)
         if not chunk_state:
-            filelog.error("Error: cannot find the file transfer id '%s'", nonl(bytestostr(chunk_id)))
+            filelog.error("Error: cannot find the file transfer id '%s'", nonl(chunk_id))
             self.send("ack-file-chunk", chunk_id, False, "file transfer id not found", chunk)
             return
         fd = chunk_state[1]
@@ -269,7 +298,7 @@ class FileTransferHandler(FileTransferAttributes):
             del self.receive_chunks_in_progress[chunk_id]
             try:
                 os.close(fd)
-            except:
+            except OSError:
                 pass
             return
         self.send("ack-file-chunk", chunk_id, True, "", chunk)
@@ -288,37 +317,54 @@ class FileTransferHandler(FileTransferAttributes):
         if written!=filesize:
             filelog.error("Error: expected a file of %i bytes, got %i", filesize, written)
             return
-        expected_digest = options.get("sha1")
+        expected_digest = options.strget("sha1")
         if expected_digest:
             self.check_digest(filename, digest.hexdigest(), expected_digest)
         start_time = chunk_state[0]
         elapsed = monotonic_time()-start_time
+        mimetype = bytestostr(mimetype)
         filelog("%i bytes received in %i chunks, took %ims", filesize, chunk, elapsed*1000)
-        t = start_thread(self.do_process_downloaded_file, "process-download", daemon=False, args=(filename, mimetype, printit, openit, filesize, options))
-        filelog("started process-download thread: %s", t)
+        self.process_downloaded_file(filename, mimetype, printit, openit, filesize, options)
 
     def accept_data(self, send_id, dtype, basefilename, printit, openit):
         #subclasses should check the flags,
         #and if ask is True, verify they have accepted this specific send_id
-        filelog("accept_data%s printing=%s, printing-ask=%s, file-transfer=%s, file-transfer-ask=%s, open-files=%s, open-files-ask=%s",
-                (send_id, dtype, basefilename, printit, openit), self.printing, self.printing_ask, self.file_transfer, self.file_transfer_ask, self.open_files, self.open_files_ask)
+        filelog("accept_data%s", (send_id, dtype, basefilename, printit, openit))
+        filelog("accept_data: printing=%s, printing-ask=%s",
+                self.printing, self.printing_ask)
+        filelog("accept_data: file-transfer=%s, file-transfer-ask=%s",
+                self.file_transfer, self.file_transfer_ask)
+        filelog("accept_data: open-files=%s, open-files-ask=%s",
+                self.open_files, self.open_files_ask)
         if printit:
-            return self.printing and not self.printing_ask
-        if not self.file_transfer or self.file_transfer_ask:
-            return False
-        if openit:
-            return self.open_files and not self.open_files_ask
-        return True
+            if not self.printing or self.printing_ask:
+                printit = False
+        elif not self.file_transfer or self.file_transfer_ask:
+            return None
+        if openit and (not self.open_files or self.open_files_ask):
+            #we can't ask in this implementation,
+            #so deny the request to open it:
+            openit = False
+        return (printit, openit)
 
     def _process_send_file(self, packet):
         #the remote end is sending us a file
+        start = monotonic_time()
         basefilename, mimetype, printit, openit, filesize, file_data, options = packet[1:8]
         send_id = ""
         if len(packet)>=9:
             send_id = packet[8]
-        if not self.accept_data(send_id, b"file", basefilename, printit, openit):
+        #basefilename should be utf8:
+        try:
+            basefilename = strtobytes(basefilename).decode("utf8")
+        except UnicodeDecodeError:
+            basefilename = bytestostr(basefilename)
+        r = self.accept_data(send_id, b"file", basefilename, printit, openit)
+        if r is None:
             filelog.warn("Warning: file transfer rejected for file '%s'", bytestostr(basefilename))
             return
+        #accept_data can override the flags:
+        printit, openit = r
         options = typedict(options)
         if printit:
             l = printlog
@@ -326,37 +372,46 @@ class FileTransferHandler(FileTransferAttributes):
         else:
             l = filelog
             assert self.file_transfer
-        l("receiving file: %s", [basefilename, mimetype, printit, openit, filesize, "%s bytes" % len(file_data), options])
+        l("receiving file: %s",
+          [basefilename, mimetype, printit, openit, filesize, "%s bytes" % len(file_data), options])
         assert filesize>0, "invalid file size: %s" % filesize
         if filesize>self.file_size_limit*1024*1024:
-            l.error("Error: file '%s' is too large:", basefilename)
+            l.error("Error: file '%s' is too large:", s(basefilename))
             l.error(" %iMB, the file size limit is %iMB", filesize//1024//1024, self.file_size_limit)
             return
-        #basefilename should be utf8:
-        try:
-            base = basefilename.decode("utf8")
-        except:
-            base = bytestostr(basefilename)
-        filename, fd = safe_open_download_file(base, mimetype)
-        self.file_descriptors.add(fd)
         chunk_id = options.strget("file-chunk-id")
+        try:
+            filename, fd = safe_open_download_file(basefilename, mimetype)
+        except OSError as e:
+            filelog("cannot save file %s / %s", basefilename, mimetype, exc_info=True)
+            filelog.error("Error: failed to save downloaded file")
+            filelog.error(" %s", e)
+            if chunk_id:
+                self.send("ack-file-chunk", chunk_id, False, "failed to create file: %s" % e, 0)
+            return
+        self.file_descriptors.add(fd)
         if chunk_id:
-            chunk_id = strtobytes(chunk_id)
-            if len(self.receive_chunks_in_progress)>=MAX_CONCURRENT_FILES:
-                self.send("ack-file-chunk", chunk_id, False, "too many file transfers in progress", 0)
+            l = len(self.receive_chunks_in_progress)
+            if l>=MAX_CONCURRENT_FILES:
+                self.send("ack-file-chunk", chunk_id, False, "too many file transfers in progress: %i" % l, 0)
                 os.close(fd)
                 return
             digest = hashlib.sha1()
             chunk = 0
             timer = self.timeout_add(CHUNK_TIMEOUT, self._check_chunk_receiving, chunk_id, chunk)
-            chunk_state = [monotonic_time(), fd, filename, mimetype, printit, openit, filesize, options, digest, 0, timer, chunk]
+            chunk_state = [
+                monotonic_time(),
+                fd, filename, mimetype,
+                printit, openit, filesize,
+                options, digest, 0, timer, chunk,
+                ]
             self.receive_chunks_in_progress[chunk_id] = chunk_state
             self.send("ack-file-chunk", chunk_id, True, "", chunk)
             return
         #not chunked, full file:
         assert file_data, "no data, got %s" % (file_data,)
         if len(file_data)!=filesize:
-            l.error("Error: invalid data size for file '%s'", basefilename)
+            l.error("Error: invalid data size for file '%s'", s(basefilename))
             l.error(" received %i bytes, expected %i bytes", len(file_data), filesize)
             return
         #check digest if present:
@@ -373,16 +428,24 @@ class FileTransferHandler(FileTransferAttributes):
             os.write(fd, file_data)
         finally:
             os.close(fd)
-        self.do_process_downloaded_file(filename, mimetype, printit, openit, filesize, options)
+        self.process_downloaded_file(filename, mimetype, printit, openit, filesize, options)
+
+
+    def process_downloaded_file(self, filename, mimetype, printit, openit, filesize, options):
+        filelog.info("downloaded %s bytes to %s file%s:",
+                     filesize, (mimetype or "temporary"), ["", " for printing"][int(printit)])
+        filelog.info(" '%s'", filename)
+        if printit or openit:
+            t = start_thread(self.do_process_downloaded_file, "process-download", daemon=False,
+                             args=(filename, mimetype, printit, openit, filesize, options))
+            filelog("started process-download thread: %s", t)
 
     def do_process_downloaded_file(self, filename, mimetype, printit, openit, filesize, options):
         filelog("do_process_downloaded_file%s", (filename, mimetype, printit, openit, filesize, options))
-        filelog.info("downloaded %s bytes to %s file%s:", filesize, (mimetype or "temporary"), ["", " for printing"][int(printit)])
-        filelog.info(" '%s'", filename)
         if printit:
             self._print_file(filename, mimetype, options)
             return
-        elif openit:
+        if openit:
             if not self.open_files:
                 filelog.warn("Warning: opening files automatically is disabled,")
                 filelog.warn(" ignoring uploaded file:")
@@ -402,13 +465,11 @@ class FileTransferHandler(FileTransferAttributes):
         from xpra.platform.printing import print_files, printing_finished, get_printers
         printers = get_printers()
         def delfile():
-            if not DELETE_PRINTER_FILE:
-                return
-            try:
-                os.unlink(filename)
-            except:
-                printlog("failed to delete print job file '%s'", filename)
-            return False
+            if DELETE_PRINTER_FILE:
+                try:
+                    os.unlink(filename)
+                except OSError:
+                    printlog("failed to delete print job file '%s'", filename)
         if not printer:
             printlog.error("Error: the printer name is missing")
             printlog.error(" printers available: %s", csv(printers.keys()) or "none")
@@ -420,7 +481,7 @@ class FileTransferHandler(FileTransferAttributes):
             delfile()
             return
         try:
-            job_options = options.get("options")
+            job_options = options.dictget("options", {})
             job_options["copies"] = copies
             job = print_files(printer, [filename], title, job_options)
         except Exception as e:
@@ -513,6 +574,10 @@ class FileTransferHandler(FileTransferAttributes):
         return True
 
 
+    def send_request_file(self, filename, openit=True):
+        self.send("request-file", filename, openit)
+
+
     def _process_open_url(self, packet):
         url, send_id = packet[1:3]
         if not self.open_url:
@@ -523,7 +588,7 @@ class FileTransferHandler(FileTransferAttributes):
             self._open_url(url)
         else:
             filelog("url '%s' not accepted", url)
-        
+
 
     def send_open_url(self, url):
         if not self.remote_open_url:
@@ -538,7 +603,7 @@ class FileTransferHandler(FileTransferAttributes):
     def do_send_open_url(self, url, send_id=""):
         self.send("open-url", url, send_id)
 
-    def send_file(self, filename, mimetype, data, filesize=0, printit=False, openit=False, options={}):
+    def send_file(self, filename, mimetype, data, filesize=0, printit=False, openit=False, options=None):
         if printit:
             l = printlog
             if not self.printing:
@@ -570,7 +635,12 @@ class FileTransferHandler(FileTransferAttributes):
                     action = "open"
         assert len(data)>=filesize, "data is smaller then the given file size!"
         data = data[:filesize]          #gio may null terminate it
-        l("send_file%s action=%s, ask=%s", (filename, mimetype, type(data), "%i bytes" % filesize, printit, openit, options), action, ask)
+        l("send_file%s action=%s, ask=%s",
+          (filename, mimetype, type(data), "%i bytes" % filesize, printit, openit, options), action, ask)
+        try:
+            filename = bytestostr(filename).encode("utf8")
+        except:
+            filename = strtobytes(filename)
         self.dump_remote_caps()
         if not self.check_file_size(action, filename, filesize):
             return False
@@ -579,24 +649,32 @@ class FileTransferHandler(FileTransferAttributes):
         self.do_send_file(filename, mimetype, data, filesize, printit, openit, options)
         return True
 
-    def send_data_request(self, action, dtype, url, mimetype="", data="", filesize=0, printit=False, openit=True, options={}):
+    def send_data_request(self, action, dtype, url, mimetype="", data="", filesize=0, printit=False, openit=True, options=None):
         send_id = uuid.uuid4().hex
         if len(self.pending_send_data)>=MAX_CONCURRENT_FILES:
             filelog.warn("Warning: %s dropped", action)
-            filelog.warn(" %i transfer%s already waiting for a response", len(self.pending_send_data), engs(self.pending_send_data))
+            filelog.warn(" %i transfer%s already waiting for a response",
+                         len(self.pending_send_data), engs(self.pending_send_data))
             return None
-        self.pending_send_data[send_id] = (dtype, url, mimetype, data, filesize, printit, openit, options)
-        self.pending_send_data_timers[send_id] = self.timeout_add(self.remote_file_ask_timeout*1000, self.send_data_ask_timeout, send_id)
-        filelog("sending data request for %s '%s' with send-id=%s", bytestostr(dtype), url, send_id)
+        self.pending_send_data[send_id] = (dtype, url, mimetype, data, filesize, printit, openit, options or {})
+        delay = self.remote_file_ask_timeout*1000
+        self.pending_send_data_timers[send_id] = self.timeout_add(delay, self.send_data_ask_timeout, send_id)
+        filelog("sending data request for %s '%s' with send-id=%s",
+                dtype, s(url), send_id)
         self.send("send-data-request", dtype, send_id, url, mimetype, filesize, printit, openit)
         return send_id
 
 
     def _process_send_data_request(self, packet):
         dtype, send_id, url, _, filesize, printit, openit = packet[1:8]
-        filelog("process send-data-request: send_id=%s, url=%s, printit=%s, openit=%s", send_id, url, printit, openit)
+        filelog("process send-data-request: send_id=%s, url=%s, printit=%s, openit=%s", s(send_id), url, printit, openit)
         def cb_answer(accept):
             filelog("accept%s=%s", (url, printit, openit), accept)
+        #filenames and url are always sent encoded as utf8:
+        try:
+            url = strtobytes(url).decode("utf8")
+        except:
+            url = bytestostr(url)
             self.send("send-data-response", send_id, accept)
         if dtype==b"file":
             if not self.file_transfer:
@@ -635,29 +713,21 @@ class FileTransferHandler(FileTransferAttributes):
 
     def _process_send_data_response(self, packet):
         send_id, accept = packet[1:3]
-        filelog("process send-data-response: send_id=%s, accept=%s", send_id, accept)
-        timer = self.pending_send_data_timers.get(send_id)
+        filelog("process send-data-response: send_id=%s, accept=%s", s(send_id), accept)
+        send_id = bytestostr(send_id)
+        timer = self.pending_send_data_timers.pop(send_id, None)
         if timer:
-            try:
-                del self.pending_send_data_timers[send_id]
-            except KeyError:
-                pass
             self.source_remove(timer)
-        v = self.pending_send_data.get(bytestostr(send_id))
-        if not v:
+        v = self.pending_send_data.pop(send_id, None)
+        if v is None:
             filelog.warn("Warning: cannot find send-file entry")
             return
-        try:
-            del self.pending_send_data[send_id]
-        except KeyError:
-            pass
         dtype = v[0]
         url = v[1]
         if accept==DENY:
-            filelog.info("the request to send %s '%s' has been denied", bytestostr(dtype), url)
+            filelog.info("the request to send %s '%s' has been denied", bytestostr(dtype), s(url))
             return
-        else:
-            assert accept in (ACCEPT, OPEN), "unknown value for send-data response: %s" % (accept,)
+        assert accept in (ACCEPT, OPEN), "unknown value for send-data response: %s" % (accept,)
         if dtype==b"file":
             mimetype, data, filesize, printit, openit, options = v[2:]
             if accept==ACCEPT:
@@ -677,37 +747,38 @@ class FileTransferHandler(FileTransferAttributes):
             filelog.error("Error: unknown datatype '%s'", dtype)
 
     def send_data_ask_timeout(self, send_id):
-        v = self.pending_send_data_timers.get(send_id)
-        if not v:
-            filelog.warn("Warning: send timeout, id '%s' not found!", send_id)
-            return False
+        v = self.pending_send_data.pop(send_id, None)
         try:
-            del self.pending_send_data[send_id]
             del self.pending_send_data_timers[send_id]
         except KeyError:
             pass
-        filename = v[0]
-        printit = v[4]
+        if not v:
+            filelog.warn("Warning: send timeout, id '%s' not found!", send_id)
+            return False
+        filename = v[1]
+        printit = v[5]
         filelog.warn("Warning: failed to %s file '%s',", ["send", "print"][printit], filename)
         filelog.warn(" the send approval request timed out")
+        return False
 
-    def do_send_file(self, filename, mimetype, data, filesize=0, printit=False, openit=False, options={}, send_id=""):
+    def do_send_file(self, filename, mimetype, data, filesize=0, printit=False, openit=False, options=None, send_id=""):
         if printit:
             action = "print"
             l = printlog
         else:
             action = "upload"
             l = filelog
-        l("do_send_file%s", (filename, mimetype, type(data), "%i bytes" % filesize, printit, openit, options))
+        l("do_send_file%s", (s(filename), mimetype, type(data), "%i bytes" % filesize, printit, openit, options))
         if not self.check_file_size(action, filename, filesize):
             return False
         u = hashlib.sha1()
         u.update(data)
         absfile = os.path.abspath(filename)
-        filelog("sha1 digest(%s)=%s", absfile, u.hexdigest())
+        filelog("sha1 digest('%s')=%s", s(absfile), u.hexdigest())
+        options = options or {}
         options["sha1"] = u.hexdigest()
         chunk_size = min(self.file_chunks, self.remote_file_chunks)
-        if chunk_size>0 and filesize>chunk_size:
+        if 0<chunk_size<filesize:
             if len(self.send_chunks_in_progress)>=MAX_CONCURRENT_FILES:
                 raise Exception("too many file transfers in progress: %i" % len(self.send_chunks_in_progress))
             #chunking is supported and the file is big enough
@@ -726,7 +797,7 @@ class FileTransferHandler(FileTransferAttributes):
         #convert str to utf8 bytes:
         try:
             base = basefilename.encode("utf8")
-        except:
+        except (AttributeError, UnicodeEncodeError):
             base = strtobytes(basefilename)
         self.send("send-file", base, mimetype, printit, openit, filesize, cdata, options, send_id)
         return True
@@ -738,19 +809,22 @@ class FileTransferHandler(FileTransferAttributes):
             chunk_state[-2] = 0         #timer has fired
             if chunk_state[-1]==chunk_no:
                 filelog.error("Error: chunked file transfer timed out on chunk %i", chunk_no)
-                del self.send_chunks_in_progress[chunk_id]
+                try:
+                    del self.send_chunks_in_progress[chunk_id]
+                except KeyError:
+                    pass
 
     def _process_ack_file_chunk(self, packet):
         #the other end received our send-file or send-file-chunk,
         #send some more file data
         filelog("ack-file-chunk: %s", packet[1:])
         chunk_id, state, error_message, chunk = packet[1:5]
+        chunk_id = bytestostr(chunk_id)
         if not state:
             filelog.error("Error: remote end is cancelling the file transfer:")
             filelog.error(" %s", error_message)
             del self.send_chunks_in_progress[chunk_id]
             return
-        chunk_id = bytestostr(chunk_id)
         chunk_state = self.send_chunks_in_progress.get(chunk_id)
         if not chunk_state:
             filelog.error("Error: cannot find the file transfer id '%s'", nonl(chunk_id))
@@ -763,7 +837,8 @@ class FileTransferHandler(FileTransferAttributes):
         if not data:
             #all sent!
             elapsed = monotonic_time()-start_time
-            filelog("%i chunks of %i bytes sent in %ims (%sB/s)", chunk, chunk_size, elapsed*1000, std_unit(chunk*chunk_size/elapsed))
+            filelog("%i chunks of %i bytes sent in %ims (%sB/s)",
+                    chunk, chunk_size, elapsed*1000, std_unit(chunk*chunk_size/elapsed))
             del self.send_chunks_in_progress[chunk_id]
             return
         assert chunk_size>0
@@ -776,3 +851,9 @@ class FileTransferHandler(FileTransferAttributes):
         timer = self.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk)
         self.send_chunks_in_progress[chunk_id] = [start_time, data, chunk_size, timer, chunk]
         self.send("send-file-chunk", chunk_id, chunk, cdata, bool(data))
+
+    def send(self, *parts):
+        raise NotImplementedError()
+
+    def compressed_wrapper(self, datatype, data, level=5):
+        raise NotImplementedError()
