@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
-# Copyright (C) 2013-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2013-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -9,24 +9,32 @@
 import socket
 import sys
 
+from xpra.os_util import WIN32
 from xpra.log import Logger
+
 log = Logger("network", "util")
 
 
-has_netifaces = True
-netifaces_version = None
-try:
-    import netifaces                #@UnresolvedImport
-    log("netifaces loaded sucessfully")
-    netifaces_version = netifaces.version        #@UndefinedVariable
-except:
-    has_netifaces = False
-    log.warn("Warning: the python netifaces package is missing")
+netifaces_version = 0
+_netifaces = None
+def import_netifaces():
+    global _netifaces, netifaces_version
+    if _netifaces is None:
+        try:
+            import netifaces                #@UnresolvedImport
+            log("netifaces loaded sucessfully")
+            _netifaces = netifaces
+            netifaces_version = netifaces.version        #@UndefinedVariable
+        except ImportError:
+            _netifaces = False
+            log.warn("Warning: the python netifaces package is missing")
+    return _netifaces
+
 iface_ipmasks = {}
 bind_IPs = None
 
 
-def get_free_tcp_port():
+def get_free_tcp_port() -> int:
     s = socket.socket()
     s.bind(('', 0))
     port = s.getsockname()[1]
@@ -35,14 +43,17 @@ def get_free_tcp_port():
 
 
 def get_interfaces():
-    if not has_netifaces:
+    netifaces = import_netifaces()
+    if not netifaces:
         return []
-    return netifaces.interfaces()           #@UndefinedVariable
+    return netifaces.interfaces()           #@UndefinedVariable pylint: disable=no-member
 
-def get_interfaces_addresses():
+def get_interfaces_addresses() -> dict:
     d = {}
-    for iface in get_interfaces():
-        d[iface] = netifaces.ifaddresses(iface)     #@UndefinedVariable
+    netifaces = import_netifaces()
+    if netifaces:
+        for iface in get_interfaces():
+            d[iface] = netifaces.ifaddresses(iface)     #@UndefinedVariable pylint: disable=no-member
     return d
 
 def get_interface(address):
@@ -59,15 +70,16 @@ def get_interface(address):
                     return iface
     return None
 
-def get_gateways():
-    if not has_netifaces:
-        return    {}
+def get_gateways() -> dict:
+    netifaces = import_netifaces()
+    if not netifaces:
+        return {}
     #versions older than 0.10.5 can crash when calling gateways()
     #https://bitbucket.org/al45tair/netifaces/issues/15/gateways-function-crash-segmentation-fault
-    if netifaces.version<'0.10.5':            #@UndefinedVariable
+    if netifaces.version<'0.10.5':            #@UndefinedVariable pylint: disable=no-member
         return {}
     try:
-        d = netifaces.gateways()            #@UndefinedVariable
+        d = netifaces.gateways()            #@UndefinedVariable pylint: disable=no-member
         AF_NAMES = {}
         for k in dir(netifaces):
             if k.startswith("AF_"):
@@ -79,14 +91,15 @@ def get_gateways():
                 continue
             gateways[AF_NAMES.get(family, family)] = gws
         return gateways
-    except:
+    except Exception:
         log("get_gateways() failed", exc_info=True)
         return {}
 
 def get_bind_IPs():
     global bind_IPs
     if not bind_IPs:
-        if has_netifaces:
+        netifaces = import_netifaces()
+        if netifaces:
             bind_IPs = do_get_bind_IPs()
         else:
             bind_IPs = ["127.0.0.1"]
@@ -95,7 +108,9 @@ def get_bind_IPs():
 def do_get_bind_IPs():
     global iface_ipmasks
     ips = []
-    ifaces = netifaces.interfaces()            #@UndefinedVariable
+    netifaces = import_netifaces()
+    assert netifaces
+    ifaces = netifaces.interfaces()            #@UndefinedVariable pylint: disable=no-member
     log("ifaces=%s", ifaces)
     for iface in ifaces:
         if_ipmasks = []
@@ -118,7 +133,9 @@ def do_get_bind_IPs():
 
 def do_get_bind_ifacemask(iface):
     ipmasks = []
-    address_types = netifaces.ifaddresses(iface)    #@UndefinedVariable
+    netifaces = import_netifaces()
+    assert netifaces
+    address_types = netifaces.ifaddresses(iface)    #@UndefinedVariable pylint: disable=no-member
     for addresses in address_types.values():
         for address in addresses:
             if 'netmask' in address and 'addr' in address:
@@ -133,7 +150,7 @@ def do_get_bind_ifacemask(iface):
     log("do_get_bind_ifacemask(%s)=%s", iface, ipmasks)
     return ipmasks
 
-def get_iface(ip):
+def get_iface(ip) -> str:
     log("get_iface(%s)", ip)
     if not ip:
         return None
@@ -176,7 +193,7 @@ def get_iface(ip):
                 log.error("incorrect ip or mask: %s/%s", test_ip, mask)
             match = True
             try:
-                for i in [0,1,2,3]:
+                for i in (0,1,2,3):
                     mask_part = int(mask_parts[i])
                     ip_part = int(ip_parts[i]) & mask_part
                     test_ip_part = int(test_ip_parts[i]) & mask_part
@@ -195,45 +212,23 @@ def get_iface(ip):
 # http://code.activestate.com/recipes/442490/
 if_nametoindex = None
 if_indextoname = None
-from xpra.os_util import WIN32, OSX
-if not WIN32:
-    library = "libc.so.6"
-    if OSX:
-        library = "/usr/lib/libc.dylib"
-    elif sys.platform.startswith("sunos"):
-        library = "libsocket.so.1"
-    elif sys.platform.startswith("freebsd"):
-        library = "/lib/libc.so.7"
-    elif sys.platform.startswith("openbsd"):
-        library = "libc.so"
-    try:
-        from ctypes import cdll, CDLL, c_char_p, c_uint, create_string_buffer
-        cdll.LoadLibrary(library)
-        #<CDLL 'libc.so.6', handle 7fcac419b000 at 7fcac1ab0c10>
-        _libc = CDLL(library)
-        log("successfully loaded socket C library from %s", library)
-    except ImportError as e:
-        log.error("library %s not found: %s", library, e)
-    except OSError as e:
-        log.error("error loading %s: %s", library, e)
-    else:
-        _libc.if_indextoname.restype = c_char_p
-        _libc.if_indextoname.argtypes = [c_uint, c_char_p]
-        _libc.if_nametoindex.restype = c_uint
-        _libc.if_nametoindex.argtypes = [c_char_p]
-        def if_nametoindex(interfaceName):
-            return _libc.if_nametoindex(create_string_buffer(interfaceName.encode()))
-        def if_indextoname(index):
-            s = create_string_buffer(b'\000' * 256)
-            return _libc.if_indextoname(c_uint(index), s)
-else:
-    def if_nametoindex(iface):
+
+if WIN32:   # pragma: no cover
+    def int_if_nametoindex(iface):
         #IPv6 addresses give us the interface as a string:
         #fe80:....%11, so try to convert "11" into 11
         try:
             return int(iface)
-        except:
+        except (TypeError, ValueError):
             return None
+    if_nametoindex = int_if_nametoindex
+else:
+    if_nametoindex = socket.if_nametoindex
+    def socket_if_indextoname(index):
+        if index<0:
+            return None
+        return socket.if_indextoname(index)
+    if_indextoname = socket_if_indextoname
 
 
 net_sys_config = None
@@ -302,7 +297,7 @@ def get_net_sys_config():
                 addproc("/proc/sys/net/ipv4/%s" % k,     "ipv4", k, int)
     return net_sys_config
 
-def get_net_config():
+def get_net_config() -> dict:
     config = {}
     try:
         from xpra.net.bytestreams import VSOCK_TIMEOUT, SOCKET_TIMEOUT, SOCKET_NODELAY
@@ -312,25 +307,27 @@ def get_net_config():
                 }
         if SOCKET_NODELAY is not None:
             config["socket.nodelay"] = SOCKET_NODELAY
-    except Exception:
+    except Exception:   # pragma: no cover
         log("get_net_config()", exc_info=True)
     return config
 
 
-def get_ssl_info():
+def get_ssl_info(show_constants=False) -> dict:
     try:
         import ssl
-    except ImportError as e:
+    except ImportError as e:    # pragma: no cover
         log("no ssl: %s", e)
         return {}
-    protocols = dict((k,int(getattr(ssl, k))) for k in dir(ssl) if k.startswith("PROTOCOL_"))
-    ops = dict((k,int(getattr(ssl, k))) for k in dir(ssl) if k.startswith("OP_"))
-    vers = dict((k,int(getattr(ssl, k))) for k in dir(ssl) if k.startswith("VERIFY_"))
-    info = {
-            "protocols"    : protocols,
-            "options"    : ops,
-            "verify"    : vers,
-            }
+    info = {}
+    if show_constants:
+        protocols = dict((k,int(getattr(ssl, k))) for k in dir(ssl) if k.startswith("PROTOCOL_"))
+        ops = dict((k,int(getattr(ssl, k))) for k in dir(ssl) if k.startswith("OP_"))
+        vers = dict((k,int(getattr(ssl, k))) for k in dir(ssl) if k.startswith("VERIFY_"))
+        info.update({
+                "protocols"    : protocols,
+                "options"    : ops,
+                "verify"    : vers,
+                })
     for k,name in {
                     "HAS_ALPN"                : "alpn",
                     "HAS_ECDH"                : "ecdh",
@@ -341,34 +338,31 @@ def get_ssl_info():
         v = getattr(ssl, k, None)
         if v is not None:
             info[name] = v
-    for k,name in {
-                    ""            : "version",
-                    "_INFO"        : "version-info",
-                    "_NUMBER"    : "version-number",
+    for k, idef in {
+                    ""           : ("version", str),
+                    "_INFO"      : ("version-info", str),
+                    "_NUMBER"    : ("version-number", int),
                     }.items():
         v = getattr(ssl, "OPENSSL_VERSION%s" % k, None)
         if v is not None:
-            info.setdefault("openssl", {})[name] = v
+            name, conv = idef
+            info.setdefault("openssl", {})[name] = conv(v)
     return info
 
 
-def get_network_caps():
-    try:
-        from xpra.platform.features import MMAP_SUPPORTED
-    except:
-        MMAP_SUPPORTED = False
-    from xpra.net.crypto import get_digests, get_crypto_caps
+def get_network_caps() -> dict:
+    from xpra.net.digest import get_digests
+    from xpra.net.crypto import get_crypto_caps
     from xpra.net.compression import get_enabled_compressors, get_compression_caps
     from xpra.net.packet_encoding import get_enabled_encoders, get_packet_encoding_caps
     digests = get_digests()
     #"hmac" is the legacy name, "xor" and "des" should not be used for salt:
-    salt_digests = [x for x in digests if x not in ("hmac", "xor", "des")]
+    salt_digests = tuple(x for x in digests if x not in ("hmac", "xor", "des"))
     caps = {
                 "digest"                : digests,
                 "salt-digest"           : salt_digests,
                 "compressors"           : get_enabled_compressors(),
                 "encoders"              : get_enabled_encoders(),
-                "mmap"                  : MMAP_SUPPORTED,
                }
     caps.update(get_crypto_caps())
     caps.update(get_compression_caps())
@@ -376,9 +370,10 @@ def get_network_caps():
     return caps
 
 
-def get_info():
+def get_info() -> dict:
     i = get_network_caps()
-    if has_netifaces:
+    netifaces = import_netifaces()
+    if netifaces:
         i["interfaces"] = get_interfaces()
         i["gateways"] = get_gateways()
     if "ssl" in sys.modules:
@@ -389,13 +384,19 @@ def get_info():
     if s:
         i["system"] = s
     i["config"] = get_net_config()
+    paramiko = sys.modules.get("paramiko")
+    if paramiko:
+        i["paramiko"] = {
+            "version"   : paramiko.__version_info__,
+            }
     return i
 
 
-def main():
+def main(): # pragma: no cover
+    from xpra.os_util import POSIX
     from xpra.util import print_nested_dict, csv
     from xpra.platform import program_context
-    from xpra.platform.netdev_query import get_interface_speed
+    from xpra.platform.netdev_query import get_interface_info
     from xpra.log import enable_color, add_debug_category, enable_debug_for
     with program_context("Network-Info", "Network Info"):
         enable_color()
@@ -406,33 +407,56 @@ def main():
             log.enable_debug()
 
         print("Network interfaces found:")
+        netifaces = import_netifaces()
         for iface in get_interfaces():
             if if_nametoindex:
-                s = "* %s (index=%s)" % (iface.ljust(20), if_nametoindex(iface))
+                print("* %s (index=%s)" % (iface.ljust(20), if_nametoindex(iface)))
             else:
-                s = "* %s" % iface
-            speed = get_interface_speed(0, iface)
-            if speed>0:
-                from xpra.simple_stats import std_unit
-                s += " (speed=%sbps)" % std_unit(speed)
-            print(s)
+                print("* %s" % iface)
+            addresses = netifaces.ifaddresses(iface)     #@UndefinedVariable pylint: disable=no-member
+            for addr, defs in addresses.items():
+                if addr in (socket.AF_INET, socket.AF_INET6):
+                    for d in defs:
+                        ip = d.get("addr")
+                        if ip:
+                            stype = {
+                                socket.AF_INET  : "IPv4",
+                                socket.AF_INET6 : "IPv6",
+                                }[addr]
+                            print(" * %s:     %s" % (stype, ip))
+                            if POSIX:
+                                from xpra.net.socket_util import create_tcp_socket
+                                try:
+                                    sock = create_tcp_socket(ip, 0)
+                                    sockfd = sock.fileno()
+                                    info = get_interface_info(sockfd, iface)
+                                    if info:
+                                        print("  %s" % info)
+                                finally:
+                                    sock.close()
+            if not POSIX:
+                info = get_interface_info(0, iface)
+                if info:
+                    print("  %s" % info)
 
+        from xpra.os_util import bytestostr
         def pver(v):
-            if type(v) in (tuple, list):
+            if isinstance(v, (tuple, list)):
                 s = ""
-                for i in range(len(v)):
-                    if i>0:
+                lastx = None
+                for x in v:
+                    if lastx is not None:
                         #dot seperated numbers
-                        if type(v[i-1])==int:
+                        if isinstance(lastx, int):
                             s += "."
                         else:
                             s += ", "
-                    s += str(v[i])
+                    s += bytestostr(x)
+                    lastx = x
                 return s
-            if type(v)==bytes:
-                from xpra.os_util import bytestostr
+            if isinstance(v, bytes):
                 v = bytestostr(v)
-            if type(v)==str and v.startswith("v"):
+            if isinstance(v, str) and v.startswith("v"):
                 return v[1:]
             return str(v)
 
@@ -440,21 +464,18 @@ def main():
         for gt,idefs in get_gateways().items():
             print("* %s" % gt)      #ie: "INET"
             for i, idef in enumerate(idefs):
-                try:
-                    if isinstance(idef, (list, tuple)):
-                        print(" [%i]           %s" % (i, csv(idef)))
-                        continue
-                except:
-                    print(" [%i]           %s" % (i, idef))
+                if isinstance(idef, (list, tuple)):
+                    print(" [%i]           %s" % (i, csv(idef)))
+                    continue
 
         print("")
         print("Protocol Capabilities:")
         netcaps = get_network_caps()
-        netif = {""    : has_netifaces}
+        netif = {""    : bool(netifaces)}
         if netifaces_version:
             netif["version"] = netifaces_version
         netcaps["netifaces"] = netif
-        print_nested_dict(netcaps)
+        print_nested_dict(netcaps, vformat=pver)
 
         print("")
         print("Network Config:")
@@ -468,7 +489,7 @@ def main():
 
         print("")
         print("SSL:")
-        print_nested_dict(get_ssl_info())
+        print_nested_dict(get_ssl_info(True))
 
         try:
             from xpra.net.crypto import crypto_backend_init, get_crypto_caps
@@ -481,7 +502,8 @@ def main():
         except Exception as e:
             print("No Crypto:")
             print(" %s" % e)
+    return 0
 
 
-if __name__ == "__main__":
+if __name__ == "__main__":  # pragma: no cover
     main()
