@@ -1,17 +1,18 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
-# Copyright (C) 2017 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 
+import dbus.service
+
 from xpra.dbus.helper import native_to_dbus
 from xpra.dbus.common import init_system_bus, init_session_bus
 from xpra.server.dbus.dbus_server import DBUS_Server_Base, INTERFACE, BUS_NAME
-import dbus.service
-
 from xpra.log import Logger
+
 log = Logger("dbus", "server")
 
 
@@ -22,11 +23,12 @@ class Proxy_DBUS_Server(DBUS_Server_Base):
             bus = init_system_bus()
         else:
             bus = init_session_bus()
-        DBUS_Server_Base.__init__(self, bus, server, BUS_NAME)
+        super().__init__(bus, server, BUS_NAME)
 
     @dbus.service.method(INTERFACE, in_signature='', out_signature='a{sv}')
     def GetInfo(self):
-        i = self.server.get_info(None)
+        #full info is available by calling get_info()
+        i = self.server.get_minimal_server_info()
         self.log(".GetInfo()=%s", i)
         try:
             v =  dbus.types.Dictionary((str(k), native_to_dbus(v)) for k,v in i.items())
