@@ -1,17 +1,19 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from threading import Thread
-from time import sleep
+import os
+from threading import Thread, Lock
 
 from xpra.server.server_core import ServerCore, get_thread_info
 from xpra.server.mixins.server_base_controlcommands import ServerBaseControlCommands
-from xpra.os_util import monotonic_time, bytestostr, strtobytes, WIN32, PYTHON3
+from xpra.server.background_worker import add_work_item
+from xpra.net.common import may_log_packet
+from xpra.os_util import monotonic_time, bytestostr, strtobytes, WIN32
 from xpra.util import (
     typedict, flatten_dict, updict, merge_dicts, envbool, envint,
     SERVER_EXIT, SERVER_ERROR, SERVER_SHUTDOWN, DETACH_REQUEST,
@@ -78,6 +80,7 @@ log("ServerBaseClass%s", SERVER_BASES)
 
 CLIENT_CAN_SHUTDOWN = envbool("XPRA_CLIENT_CAN_SHUTDOWN", True)
 INIT_THREAD_TIMEOUT = envint("XPRA_INIT_THREAD_TIMEOUT", 10)
+MDNS_CLIENT_COUNT = envbool("XPRA_MDNS_CLIENT_COUNT", True)
 
 
 """
@@ -97,12 +100,15 @@ class ServerBase(ServerBaseClass):
         self._authenticated_packet_handlers = {}
         self._authenticated_ui_packet_handlers = {}
 
+        self.display_pid = 0
         self._server_sources = {}
         self.client_properties = {}
         self.ui_driver = None
         self.sharing = None
         self.lock = None
         self.init_thread = None
+        self.init_thread_callbacks = []
+        self.init_thread_lock = Lock()
 
         self.idle_timeout = 0
         #duplicated from Server Source...
@@ -129,26 +135,31 @@ class ServerBase(ServerBaseClass):
         if self.dbus_server:
             self.dbus_server.Event(str(args[0]), [str(x) for x in args[1:]])
 
+    def get_server_source(self, proto):
+        return self._server_sources.get(proto)
+
 
     def init(self, opts):
         #from now on, use the logger for parsing errors:
         from xpra.scripts import config
         config.warn = log.warn
         for c in SERVER_BASES:
+            start = monotonic_time()
             c.init(self, opts)
-        log("ServerBase.init(%s)", opts)
-
+            end = monotonic_time()
+            log("%3ims in %s.init", 1000*(end-start), c)
         self.sharing = opts.sharing
         self.lock = opts.lock
         self.idle_timeout = opts.idle_timeout
-        self.av_sync = opts.av_sync
         self.bandwidth_detection = opts.bandwidth_detection
 
-
     def setup(self):
         log("starting component init")
         for c in SERVER_BASES:
+            start = monotonic_time()
             c.setup(self)
+            end = monotonic_time()
+            log("%3ims in %s.setup", 1000*(end-start), c)
         self.init_thread = Thread(target=self.threaded_init)
         self.init_thread.start()
 
@@ -158,14 +169,38 @@ class ServerBase(ServerBaseClass):
         threaded_server_init()
         for c in SERVER_BASES:
             if c!=ServerCore:
-                c.threaded_setup(self)
+                try:
+                    c.threaded_setup(self)
+                except Exception:
+                    log.error("Error during threaded setup of %s", c, exc_info=True)
+        #populate the platform info cache:
+        from xpra.version_util import get_platform_info
+        get_platform_info()
+        with self.init_thread_lock:
+            for cb in self.init_thread_callbacks:
+                try:
+                    cb()
+                except Exception as e:
+                    log("threaded_init()", exc_info=True)
+                    log.error("Error in initialization thread callback %s", cb)
+                    log.error(" %s", e)
         log("threaded_init() end")
 
+    def after_threaded_init(self, callback):
+        with self.init_thread_lock:
+            if not self.init_thread or not self.init_thread.is_alive():
+                callback()
+            else:
+                self.init_thread_callbacks.append(callback)
+
     def wait_for_threaded_init(self):
-        assert self.init_thread
+        if not self.init_thread:
+            #looks like we didn't make it as far as calling setup()
+            log("wait_for_threaded_init() no init thread")
+            return
         log("wait_for_threaded_init() %s.is_alive()=%s", self.init_thread, self.init_thread.is_alive())
         if self.init_thread.is_alive():
-            log.info("waiting for video encoders initialization")
+            log.info("waiting for initialization thread to complete")
             self.init_thread.join(INIT_THREAD_TIMEOUT)
             if self.init_thread.is_alive():
                 log.warn("Warning: initialization thread is still active")
@@ -178,6 +213,7 @@ class ServerBase(ServerBaseClass):
 
     def do_cleanup(self):
         self.server_event("exit")
+        self.wait_for_threaded_init()
         self.cancel_mp3_stream_check_timer()
         for c in SERVER_BASES:
             if c!=ServerCore:
@@ -200,6 +236,13 @@ class ServerBase(ServerBaseClass):
         self.timeout_add(500, self.clean_quit)
 
 
+    def get_mdns_info(self) -> dict:
+        mdns_info = ServerCore.get_mdns_info(self)
+        if MDNS_CLIENT_COUNT:
+            mdns_info["clients"] = len(self._server_sources)
+        return mdns_info
+
+
     ######################################################################
     # handle new connections:
     def handle_sharing(self, proto, ui_client=True, detach_request=False, share=False, uuid=None):
@@ -207,7 +250,9 @@ class ServerBase(ServerBaseClass):
         disconnected = 0
         existing_sources = set(ss for p,ss in self._server_sources.items() if p!=proto)
         is_existing_client = uuid and any(ss.uuid==uuid for ss in existing_sources)
-        log("handle_sharing%s lock=%s, sharing=%s, existing sources=%s, is existing client=%s", (proto, ui_client, detach_request, share, uuid), self.lock, self.sharing, existing_sources, is_existing_client)
+        log("handle_sharing%s lock=%s, sharing=%s, existing sources=%s, is existing client=%s",
+            (proto, ui_client, detach_request, share, uuid),
+            self.lock, self.sharing, existing_sources, is_existing_client)
         #if other clients are connected, verify we can steal or share:
         if existing_sources and not is_existing_client:
             if self.sharing is True or (self.sharing is None and share and all(ss.share for ss in existing_sources)):
@@ -222,7 +267,7 @@ class ServerBase(ServerBaseClass):
             if detach_request and p!=proto:
                 self.disconnect_client(p, DETACH_REQUEST)
                 disconnected += 1
-            elif uuid and ss.uuid==uuid:
+            elif uuid and ss.uuid==uuid and ui_client and ss.ui_client:
                 self.disconnect_client(p, NEW_CLIENT, "new connection from the same uuid")
                 disconnected += 1
             elif ui_client and ss.ui_client:
@@ -314,10 +359,11 @@ class ServerBase(ServerBaseClass):
                 self.ydpi = c.intget("dpi.y", 0)
                 self.double_click_time = c.intget("double_click.time", -1)
                 self.double_click_distance = c.intpair("double_click.distance", (-1, -1))
-                self.antialias = c.dictget("antialias")
+                self.antialias = c.dictget("antialias", {})
                 self.cursor_size = c.intget("cursor.size", 0)
             #FIXME: this belongs in DisplayManager!
-            screenlog("dpi=%s, dpi.x=%s, dpi.y=%s, antialias=%s, cursor_size=%s", self.dpi, self.xdpi, self.ydpi, self.antialias, self.cursor_size)
+            screenlog("dpi=%s, dpi.x=%s, dpi.y=%s, antialias=%s, cursor_size=%s",
+                      self.dpi, self.xdpi, self.ydpi, self.antialias, self.cursor_size)
             log("double-click time=%s, distance=%s", self.double_click_time, self.double_click_distance)
             #if we're not sharing, reset all the settings:
             reset = share_count==0
@@ -325,12 +371,12 @@ class ServerBase(ServerBaseClass):
 
         self.accept_client(proto, c)
         #use blocking sockets from now on:
-        if not (PYTHON3 and WIN32):
+        if not WIN32:
             set_socket_timeout(proto._conn, None)
 
         def drop_client(reason="unknown", *args):
             self.disconnect_client(proto, reason, *args)
-        cc_class = self.get_client_connection_class()
+        cc_class = self.get_client_connection_class(c)
         ss = cc_class(proto, drop_client,
                       self.session_name, self,
                       self.idle_add, self.timeout_add, self.source_remove,
@@ -345,23 +391,25 @@ class ServerBase(ServerBaseClass):
             #close it already
             ss.close()
             raise
-        self.notify_new_user(ss)
         self._server_sources[proto] = ss
+        add_work_item(self.mdns_update)
         #process ui half in ui thread:
         send_ui = ui_client and not is_request
         self.idle_add(self._process_hello_ui, ss, c, auth_caps, send_ui, share_count)
 
-    def get_client_connection_class(self):
-        from xpra.server.source.client_connection import ClientConnection
-        return ClientConnection
+    def get_client_connection_class(self, caps):
+        from xpra.server.source.client_connection_factory import get_client_connection_class
+        return get_client_connection_class(caps)
 
 
-    def _process_hello_ui(self, ss, c, auth_caps, send_ui, share_count):
+    def _process_hello_ui(self, ss, c, auth_caps, send_ui : bool, share_count : int):
         #adds try:except around parse hello ui code:
         try:
             if self._closing:
                 raise Exception("server is shutting down")
 
+            self.notify_new_user(ss)
+
             self.parse_hello(ss, c, send_ui)
             #send_hello will take care of sending the current and max screen resolutions
             root_w, root_h = self.get_root_window_size()
@@ -399,7 +447,7 @@ class ServerBase(ServerBaseClass):
     def client_startup_complete(self, ss):
         ss.startup_complete()
         self.server_event("startup-complete", ss.uuid)
-        if not self.start_after_connect_done:
+        if not self.start_after_connect_done:   #pylint: disable=access-member-before-definition
             self.start_after_connect_done = True
             self.exec_after_connect_commands()
         self.exec_on_connect_commands()
@@ -408,7 +456,8 @@ class ServerBase(ServerBaseClass):
         server_uuid = c.strget("server_uuid")
         if server_uuid:
             if server_uuid==self.uuid:
-                self.send_disconnect(proto, "cannot connect a client running on the same display that the server it connects to is managing - this would create a loop!")
+                self.send_disconnect(proto, "cannot connect a client running on the same display"
+                                     +" that the server it connects to is managing - this would create a loop!")
                 return  False
             log.warn("This client is running within the Xpra server %s", server_uuid)
         return True
@@ -423,22 +472,16 @@ class ServerBase(ServerBaseClass):
     def get_server_features(self, server_source=None):
         #these are flags that have been added over time with new versions
         #to expose new server features:
-        f = dict((k, True) for k in (
-                #all these flags are assumed enabled in 0.17 (they are present in 0.14.x onwards):
-                "toggle_cursors_bell_notify",
-                "toggle_keyboard_sync",
-                "xsettings-tuple",
-                "event_request",
-                "notify-startup-complete",
-                "server-events",
-                ))
+        f = {
+            "toggle_keyboard_sync" : True,  #v4.0 clients assume this is always available
+            }
         for c in SERVER_BASES:
             if c!=ServerCore:
                 merge_dicts(f, c.get_server_features(self, server_source))
         return f
 
     def make_hello(self, source):
-        capabilities = ServerCore.make_hello(self, source)
+        capabilities = super().make_hello(source)
         for c in SERVER_BASES:
             if c!=ServerCore:
                 merge_dicts(capabilities, c.get_caps(self, source))
@@ -446,6 +489,7 @@ class ServerBase(ServerBaseClass):
         if source.wants_display:
             capabilities.update({
                  "max_desktop_size"             : self.get_max_screen_size(),
+                 "display"                      : os.environ.get("DISPLAY", "Main"),
                  })
         if source.wants_features:
             capabilities.update({
@@ -459,33 +503,50 @@ class ServerBase(ServerBaseClass):
                  "pointer"                      : server_features.input_devices,
                  })
             capabilities.update(flatten_dict(self.get_server_features(source)))
-        #this is a feature, but we would need the hello request
-        #to know if it is really needed.. so always include it:
-        capabilities["exit_server"] = True
+        capabilities["configure.pointer"] = True    #v4 clients assume this is enabled
         return capabilities
 
     def send_hello(self, server_source, root_w, root_h, server_cipher):
         capabilities = self.make_hello(server_source)
-        if server_source.wants_encodings:
+        if server_source.wants_encodings and server_features.windows:
             try:
                 from xpra.codecs.loader import codec_versions
             except ImportError:
                 log("no codecs", exc_info=True)
             else:
-                updict(capabilities, "encoding", codec_versions, "version")
-                for k,v in self.get_encoding_info().items():
-                    if k=="":
-                        k = "encodings"
-                    else:
-                        k = "encodings.%s" % k
-                    capabilities[k] = v
+                def add_encoding_caps(d):
+                    updict(d, "encoding", codec_versions, "version")
+                    for k,v in self.get_encoding_info().items():
+                        if k=="":
+                            k = "encodings"
+                        else:
+                            k = "encodings.%s" % k
+                        d[k] = v
+                if server_source.encodings_packet:
+                    #we can send it later,
+                    #when the init thread has finished:
+                    def send_encoding_caps():
+                        d = {}
+                        add_encoding_caps(d)
+                        #make sure the 'hello' packet goes out first:
+                        self.idle_add(server_source.send_async, "encodings", d)
+                    self.after_threaded_init(send_encoding_caps)
+                else:
+                    self.wait_for_threaded_init()
+                    add_encoding_caps(capabilities)
+                #check for mmap:
+                if getattr(self, "mmap_size", 0)==0:
+                    self.after_threaded_init(server_source.print_encoding_info)
         if server_source.wants_display:
             capabilities.update({
                          "actual_desktop_size"  : (root_w, root_h),
                          "root_window_size"     : (root_w, root_h),
                          })
-        if self._reverse_aliases and server_source.wants_aliases:
-            capabilities["aliases"] = self._reverse_aliases
+        if self._aliases and server_source.wants_aliases:
+            reverse_aliases = {}
+            for i, packet_type in self._aliases.items():
+                reverse_aliases[packet_type] = i
+            capabilities["aliases"] = reverse_aliases
         if server_cipher:
             capabilities.update(server_cipher)
         server_source.send_hello(capabilities)
@@ -496,7 +557,7 @@ class ServerBase(ServerBaseClass):
     def _process_info_request(self, proto, packet):
         log("process_info_request(%s, %s)", proto, packet)
         #ignoring the list of client uuids supplied in packet[1]
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if not ss:
             return
         categories = None
@@ -511,15 +572,17 @@ class ServerBase(ServerBaseClass):
             ss.send_info_response(info)
         self.get_all_info(info_callback, proto, None)
 
-    def send_hello_info(self, proto, flatten=True):
+    def send_hello_info(self, proto):
+        self.wait_for_threaded_init()
         start = monotonic_time()
         def cb(proto, info):
-            self.do_send_info(proto, info, flatten)
+            self.do_send_info(proto, info)
             end = monotonic_time()
-            log.info("processed %s info request from %s in %ims", ["structured", "flat"][flatten], proto._conn, (end-start)*1000)
+            log.info("processed info request from %s in %ims",
+                     proto._conn, (end-start)*1000)
         self.get_all_info(cb, proto, None)
 
-    def get_ui_info(self, proto, client_uuids=None, *args):
+    def get_ui_info(self, proto, client_uuids=None, *args) -> dict:
         """ info that must be collected from the UI thread
             (ie: things that query the display)
         """
@@ -531,11 +594,11 @@ class ServerBase(ServerBaseClass):
                 log.error("Error gathering UI info on %s", c, exc_info=True)
         return info
 
-    def get_thread_info(self, proto):
+    def get_thread_info(self, proto) -> dict:
         return get_thread_info(proto, tuple(self._server_sources.keys()))
 
 
-    def get_info(self, proto=None, client_uuids=None):
+    def get_info(self, proto=None, client_uuids=None) -> dict:
         log("ServerBase.get_info%s", (proto, client_uuids))
         start = monotonic_time()
         info = ServerCore.get_info(self, proto)
@@ -557,7 +620,7 @@ class ServerBase(ServerBaseClass):
         log("ServerBase.get_info took %.1fms", 1000.0*(monotonic_time()-start))
         return info
 
-    def get_packet_handlers_info(self):
+    def get_packet_handlers_info(self) -> dict:
         info = ServerCore.get_packet_handlers_info(self)
         info.update({
             "authenticated" : sorted(self._authenticated_packet_handlers.keys()),
@@ -566,7 +629,7 @@ class ServerBase(ServerBaseClass):
         return info
 
 
-    def get_features_info(self):
+    def get_features_info(self) -> dict:
         i = {
              "sharing"          : self.sharing is not False,
              "idle_timeout"     : self.idle_timeout,
@@ -574,7 +637,7 @@ class ServerBase(ServerBaseClass):
         i.update(self.get_server_features())
         return i
 
-    def do_get_info(self, proto, server_sources=None):
+    def do_get_info(self, proto, server_sources=None) -> dict:
         start = monotonic_time()
         info = {}
         def up(prefix, d):
@@ -597,8 +660,10 @@ class ServerBase(ServerBaseClass):
             })
 
         # other clients:
-        info["clients"] = {""                   : len([p for p in self._server_sources.keys() if p!=proto]),
-                           "unauthenticated"    : len([p for p in self._potential_protocols if ((p is not proto) and (p not in self._server_sources.keys()))])}
+        info["clients"] = {
+            ""                   : sum(1 for p in self._server_sources if p!=proto),
+            "unauthenticated"    : sum(1 for p in self._potential_protocols if ((p is not proto) and (p not in self._server_sources))),
+           }
         #find the server source to report on:
         n = len(server_sources or [])
         if n==1:
@@ -625,7 +690,7 @@ class ServerBase(ServerBaseClass):
         Allows us to keep window properties for a client after disconnection.
         (we keep it in a map with the client's uuid as key)
         """
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             ss.set_client_properties(wid, window, typedict(new_client_properties))
             #filter out encoding properties, which are expected to be set everytime:
@@ -655,13 +720,13 @@ class ServerBase(ServerBaseClass):
         log("client has requested compression level=%s", level)
         proto.set_compression_level(level)
         #echo it back to the client:
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             ss.set_deflate(level)
 
     def _process_sharing_toggle(self, proto, packet):
         assert self.sharing is None
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if not ss:
             return
         sharing = bool(packet[1])
@@ -670,11 +735,12 @@ class ServerBase(ServerBaseClass):
             #disconnect other users:
             for p,ss in tuple(self._server_sources.items()):
                 if p!=proto:
-                    self.disconnect_client(p, DETACH_REQUEST, "client %i no longer wishes to share the session" % ss.counter)
+                    self.disconnect_client(p, DETACH_REQUEST,
+                                           "client %i no longer wishes to share the session" % ss.counter)
 
     def _process_lock_toggle(self, proto, packet):
         assert self.lock is None
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             ss.lock = bool(packet[1])
             log("lock set to %s for client %i", ss.lock, ss.counter)
@@ -684,12 +750,12 @@ class ServerBase(ServerBaseClass):
 
     ######################################################################
     # http server and http audio stream:
-    def get_http_info(self):
+    def get_http_info(self) -> dict:
         info = ServerCore.get_http_info(self)
         info["clients"] = len(self._server_sources)
         return info
 
-    def get_http_scripts(self):
+    def get_http_scripts(self) -> dict:
         scripts = ServerCore.get_http_scripts(self)
         scripts["/audio.mp3"] = self.http_audio_mp3_request
         return scripts
@@ -700,7 +766,7 @@ class ServerBase(ServerBaseClass):
             return None
         try:
             args_str = handler.path.split("?", 1)[1]
-        except:
+        except IndexError:
             return err()
         #parse args:
         args = {}
@@ -723,12 +789,26 @@ class ServerBase(ServerBaseClass):
         if not source:
             httplog.warn("Warning: no client matching uuid '%s'", uuid)
             return err()
+        #don't close the connection when handler.finish() is called,
+        #we will continue to write to this socket as we process more buffers:
+        finish = handler.finish
+        def do_finish():
+            try:
+                finish()
+            except:
+                log("error calling %s", finish, exc_info=True)
+        def noop():
+            pass
+        handler.finish = noop
         state = {}
-        def new_buffer(_sound_source, data, _metadata, packet_metadata=[]):
+        def new_buffer(_sound_source, data, _metadata, packet_metadata=()):
+            if state.get("failed"):
+                return
             if not state.get("started"):
                 httplog.warn("buffer received but stream is not started yet")
-                err()
                 source.stop_sending_sound()
+                err()
+                do_finish()
                 return
             count = state.get("buffers", 0)
             httplog("new_buffer [%i] for %s sound stream: %i bytes", count, state.get("codec", "?"), len(data))
@@ -738,29 +818,45 @@ class ServerBase(ServerBaseClass):
                 for x in packet_metadata:
                     handler.wfile.write(x)
                 handler.wfile.write(data)
+                handler.wfile.flush()
             except Exception as e:
+                state["failed"] = True
+                httplog("failed to send new audio buffer", exc_info=True)
                 httplog.warn("Error: failed to send audio packet:")
                 httplog.warn(" %s", e)
                 source.stop_sending_sound()
-                return
-        def new_stream(_sound_source, codec):
+                do_finish()
+        def new_stream(sound_source, codec):
+            codec = bytestostr(codec)
             httplog("new_stream: %s", codec)
-            state["started"] = True
-            state["buffers"] = 0
-            state["codec"] = codec
-            handler.send_response(200)
+            sound_source.codec = codec
             headers = {
                 "Content-type"      : "audio/mpeg",
                 }
-            for k,v in headers.items():
-                handler.send_header(k, v)
-            handler.end_headers()
+            try:
+                handler.send_response(200)
+                for k,v in headers.items():
+                    handler.send_header(k, v)
+                handler.end_headers()
+            except ValueError:
+                httplog("new_stream error writing headers", exc_info=True)
+                state["failed"] = True
+                source.stop_sending_sound()
+                do_finish()
+            else:
+                state["started"] = True
+                state["buffers"] = 0
+                state["codec"] = codec
         def timeout_check():
             if not state.get("started"):
                 err()
+                source.stop_sending_sound()
         if source.sound_source:
             source.stop_sending_sound()
-        source.start_sending_sound("mp3", volume=1.0, new_stream=new_stream, new_buffer=new_buffer, skip_client_codec_check=True)
+        def start_sending_sound():
+            source.start_sending_sound("mp3", volume=1.0, new_stream=new_stream,
+                                       new_buffer=new_buffer, skip_client_codec_check=True)
+        self.idle_add(start_sending_sound)
         self.mp3_stream_check_timer = self.timeout_add(1000*5, timeout_check)
 
     def cancel_mp3_stream_check_timer(self):
@@ -787,6 +883,7 @@ class ServerBase(ServerBaseClass):
         source = self._server_sources.pop(protocol, None)
         if source:
             self.cleanup_source(source)
+            add_work_item(self.mdns_update)
         for c in SERVER_BASES:
             c.cleanup_protocol(self, protocol)
         return source
@@ -809,18 +906,17 @@ class ServerBase(ServerBaseClass):
     def last_client_exited(self):
         #must run from the UI thread (modifies focus and keys)
         netlog("last_client_exited() exit_with_client=%s", self.exit_with_client)
+        self.reset_server_timeout(True)
+        for c in SERVER_BASES:
+            if c!=ServerCore:
+                try:
+                    c.last_client_exited(self)
+                except Exception:
+                    log("last_client_exited calling %s", c.last_client_exited, exc_info=True)
         if self.exit_with_client:
             if not self._closing:
                 netlog.info("Last client has disconnected, terminating")
                 self.clean_quit(False)
-        else:
-            self.reset_server_timeout(True)
-            for c in SERVER_BASES:
-                if c!=ServerCore:
-                    try:
-                        c.last_client_exited(self)
-                    except:
-                        log("last_client_exited calling %s", c.last_client_exited, exc_info=True)
 
 
     def set_ui_driver(self, source):
@@ -842,7 +938,7 @@ class ServerBase(ServerBaseClass):
                 c.reset_focus(self)
 
 
-    def get_all_protocols(self):
+    def get_all_protocols(self) -> list:
         return list(self._potential_protocols) + list(self._server_sources.keys())
 
 
@@ -855,7 +951,7 @@ class ServerBase(ServerBaseClass):
     def _log_disconnect(self, proto, *args):
         #skip logging of disconnection events for server sources
         #we have tagged during hello ("info_request", "exit_request", etc..)
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss and not ss.log_disconnect:
             #log at debug level only:
             netlog(*args)
@@ -870,21 +966,33 @@ class ServerBase(ServerBaseClass):
 
     ######################################################################
     # packets:
+    def add_packet_handlers(self, defs, main_thread=True):
+        for packet_type, handler in defs.items():
+            self.add_packet_handler(packet_type, handler, main_thread)
+
+    def add_packet_handler(self, packet_type, handler, main_thread=True):
+        netlog("add_packet_handler%s", (packet_type, handler, main_thread))
+        if main_thread:
+            handlers = self._authenticated_ui_packet_handlers
+        else:
+            handlers = self._authenticated_packet_handlers
+        handlers[packet_type] = handler
+
     def init_packet_handlers(self):
         for c in SERVER_BASES:
             c.init_packet_handlers(self)
-        self._authenticated_packet_handlers.update({
-            "sharing-toggle":                       self._process_sharing_toggle,
-            "lock-toggle":                          self._process_lock_toggle,
-          })
-        self._authenticated_ui_packet_handlers.update({
-            #attributes / settings:
-            "server-settings":                      self._process_server_settings,
-            "set_deflate":                          self._process_set_deflate,
-            #requests:
-            "shutdown-server":                      self._process_shutdown_server,
-            "exit-server":                          self._process_exit_server,
-            "info-request":                         self._process_info_request,
+        #no need for main thread:
+        self.add_packet_handlers({
+            "sharing-toggle"    : self._process_sharing_toggle,
+            "lock-toggle"       : self._process_lock_toggle,
+            }, False)
+        #attributes / settings:
+        self.add_packet_handlers({
+            "server-settings"   : self._process_server_settings,
+            "set_deflate"       : self._process_set_deflate,
+            "shutdown-server"   : self._process_shutdown_server,
+            "exit-server"       : self._process_exit_server,
+            "info-request"      : self._process_info_request,
             })
 
     def init_aliases(self):
@@ -897,33 +1005,34 @@ class ServerBase(ServerBaseClass):
         try:
             handler = None
             packet_type = bytestostr(packet[0])
+            def call_handler():
+                may_log_packet(False, packet_type, packet)
+                handler(proto, packet)
             if proto in self._server_sources:
                 handler = self._authenticated_ui_packet_handlers.get(packet_type)
                 if handler:
                     netlog("process ui packet %s", packet_type)
-                    self.idle_add(handler, proto, packet)
+                    self.idle_add(call_handler)
                     return
                 handler = self._authenticated_packet_handlers.get(packet_type)
                 if handler:
                     netlog("process non-ui packet %s", packet_type)
-                    handler(proto, packet)
+                    call_handler()
                     return
             handler = self._default_packet_handlers.get(packet_type)
             if handler:
                 netlog("process default packet %s", packet_type)
-                handler(proto, packet)
+                call_handler()
                 return
             def invalid_packet():
-                ss = self._server_sources.get(proto)
-                if not self._closing and not proto._closed and (ss is None or not ss.is_closed()):
+                ss = self.get_server_source(proto)
+                if not self._closing and not proto.is_closed() and (ss is None or not ss.is_closed()):
                     netlog("invalid packet: %s", packet)
                     netlog.error("Error: unknown or invalid packet type '%s'", packet_type)
                     netlog.error(" received from %s", proto)
                 if not ss:
                     proto.close()
             self.idle_add(invalid_packet)
-        except KeyboardInterrupt:
-            raise
         except Exception:
             netlog.error("Error processing a '%s' packet", packet_type)
             netlog.error(" received from %s:", proto)
